import 'package:nexbounce/nexbounce.dart';

typedef Subscriber<T> = void Function(T state);

class Subscription<T> {
  final Store<T> store;
  final Subscriber<T> subscriber;

  const Subscription({required this.store, required this.subscriber});

  void cancel() => store._subscribers.remove(subscriber);
}

/// A base class that has to be extended to create Actions.
abstract class Action<T> {
  const Action();

  T activate(T state);
}

/// A base class that has to be extended to create stores.
abstract class Store<T> {
  final _subscribers = <Subscriber<T>>{};
  final _debouncer = Nexbounce();
  final bool logger;

  T _state;

  Store(this._state, {this.logger = false});

  /// Returns the current state of store.
  T get state => _state;

  /// Dispatches an action to update the state.
  void dispatch(Action<T> action) {
    final newState = action.activate(state);

    if (newState != state) {
      if (logger) {
        _log(
          oldState: state,
          newState: newState,
          action: action,
        );
      }

      _state = newState;
      _publish();
    }
  }

  /// Accepts a callback as its argument which is run after state change.
  /// Note that running callbacks is debounced.
  Subscription<T> subscribe(Subscriber<T> subscriber) {
    _subscribers.add(subscriber);
    return Subscription(store: this, subscriber: subscriber);
  }

  /// Same as `subscribe` but callback is immediately called for the first time.
  Subscription<T> runAndSubscribe(Subscriber<T> subscriber) {
    subscriber(state);
    return subscribe(subscriber);
  }

  void _publish() {
    _debouncer.enqueue(() {
      for (final subscriber in _subscribers) {
        subscriber(state);
      }
    });
  }

  void _log({
    required T oldState,
    required T newState,
    required Action<T> action,
  }) =>
      print(
        '''
store: $this
action: $action
oldState: $oldState
newState: $newState
''',
      );
}
